/*	Experimental TMD toon shading example by Lameguy64
	2014 Meido-Tek Productions
	
	Demonstrates:
		- Everything from tmdtest.
		- Outline rendering for toon-shaded 3D graphics.
		
	Controls:
		D-Pad			- Look around.
		Triangle/Cross 	- Move forward/backward.
		Square/Circle	- Strafe left/right.
		
	To compile:
		ccpsx -Xo$80010000 -O3 tmdtoon.c -otmdtoon.cpe
		cpe2x tmdtoon.cpe
		
	
	Outline rendering was made possible on the PlayStation by simply drawing 2 copies of
	the same model, one that is completely black (outline model), and the other
	colored/textured (fill model). Both models are rendered like this:
		- Draw the 'outline' model twice, one slightly larger and the other slightly smaller.
		- Draw the 'fill' model over the 'outline model with OT_LENGTH-1.

	For best results, the axis point of the model must be at the exact center. However, the
	outline result may not be as accurate as what other toon-shaded games had but its acceptable
	for what the PlayStation can handle.
	
*/

#include <sys/types.h>
#include <libetc.h>
#include <libgte.h>
#include <libgpu.h>
#include <libgs.h>

#include "platform.c"
#include "twilight.c"

// Maximum number of objects
#define MAX_OBJECTS 3


// Screen resolution and dither mode
#define SCREEN_XRES		640
#define SCREEN_YRES 	480
#define DITHER			0

#define CENTERX			SCREEN_XRES/2
#define CENTERY			SCREEN_YRES/2


// Increasing this value (max is 14) reduces sorting errors in certain cases
#define OT_LENGTH	14

#define OT_ENTRIES	1<<OT_LENGTH
#define PACKETMAX	2048

#define SUBOT_LENGTH	12
#define SUBOT_ENTRIES	1<<SUBOT_LENGTH


GsOT		myOT[2];						// OT handlers
GsOT_TAG	myOT_TAG[2][OT_ENTRIES];		// OT tables
PACKET		myPacketArea[2][PACKETMAX*24];	// Packet buffers
int			myActiveBuff=0;					// Page index counter


// Camera coordinates
struct {
	int		x,y,z;
	int		pan,til,rol;
	VECTOR	pos;
	SVECTOR rot;
	GsRVIEW2 view;
	GsCOORDINATE2 coord2;
} Camera={0};


// Object handler
GsDOBJ2	Object[MAX_OBJECTS]={0};
int		ObjectCount=0;

// Lighting coordinates
GsF_LIGHT pslt[1];


// Prototypes
int main();

void CalculateCamera();
void PutToonObjectOut(VECTOR pos, SVECTOR rot, GsDOBJ2 *obj);
void PutToonObject(VECTOR pos, SVECTOR rot, GsDOBJ2 *obj);
void PutObject(VECTOR pos, SVECTOR rot, GsDOBJ2 *obj);

int LinkModel(u_long *tmd, GsDOBJ2 *obj);
void LoadTexture(u_long *addr);

void init();
void PrepDisplay();
void Display();


// Main stuff
int main() {
	
	int PadStatus;
	int Accel;
	
	// Object coordinates
	VECTOR	plat_pos={0};
	SVECTOR	plat_rot={0};
	
	VECTOR	obj_pos={0};
	SVECTOR	obj_rot={0};
	
	VECTOR	bulb_pos={0};
	
	
	// Player coordinates
	struct {
		int x,xv;
		int y,yv;
		int z,zv;
		int pan,panv;
		int til,tilv;
	} Player = {0};
	
	
	// Init everything
	init();
	
	
	// Load the textures and models
	LoadTexture((u_long*)tim_twilight);
	ObjectCount += LinkModel((u_long*)tmd_platform, &Object[0]);
	ObjectCount += LinkModel((u_long*)tmd_twilight_f, &Object[1]);
	ObjectCount += LinkModel((u_long*)tmd_twilight_b, &Object[2]);
	
	Object[0].attribute |= GsDIV2;	// Set x2 subdivision to the platform to reduce clipping errors
	Object[1].attribute ^= GsLOFF; // Re-enable lighting for the pony model
	
	
	// Set camera/player positions
	Player.x = ONE*-640;
	Player.y = ONE*510;
	Player.z = ONE*800;
	
	Player.pan = -660;
	Player.til = -245;
	
	
	// Set object positions
	plat_pos.vy = 1024;
	obj_pos.vy = 0;
	
	bulb_pos.vz = -800;
	bulb_pos.vy = -400;
	
	
	while(1) {
		
		PadStatus = PadRead(0);
		
		
		// Player look
		if (PadStatus & PADLup)		Player.tilv += 4;
		if (PadStatus & PADLdown)	Player.tilv -= 4;
		if (PadStatus & PADLleft)	Player.panv -= 4;
		if (PadStatus & PADLright)	Player.panv += 4;
	
		// Player movement
		if (PadStatus & PADRup) {	// Forward
			Accel = ccos(Player.til);
			Player.zv -= ((ccos(Player.pan)*Accel)/ONE)*2;
			Player.xv -= ((csin(Player.pan)*Accel)/ONE)*2;
			Player.yv += (csin(Player.til)*2);
		}

		if (PadStatus & PADRdown) {	// Backward
			Accel = ccos(Player.til);
			Player.zv += ((ccos(Player.pan)*Accel)/ONE)*2;
			Player.xv += ((csin(Player.pan)*Accel)/ONE)*2;
			Player.yv -= (csin(Player.til)*2);
		}
		
		if (PadStatus & PADRleft) { // Strafe left
			Player.zv -= (csin(Player.pan)*2);
			Player.xv += (ccos(Player.pan)*2);
		}

		if (PadStatus & PADRright) { // Strafe right
			Player.zv += (csin(Player.pan)*2);
			Player.xv -= (ccos(Player.pan)*2);
		}
		
		
		// Calculate player coordinates according to the velocities
		Player.x += Player.xv;
		Player.y += Player.yv;
		Player.z += Player.zv;
		Player.pan += Player.panv;
		Player.til += Player.tilv;
		
		Player.xv *=0.9f;
		Player.yv *=0.9f;
		Player.zv *=0.9f;
		Player.panv *=0.9f;
		Player.tilv *=0.9f;
		
		
		// Translate player coordinates for the camera
		Camera.pos.vx = Player.x/ONE;
		Camera.pos.vy = Player.y/ONE;
		Camera.pos.vz = Player.z/ONE;
		Camera.rot.vy = -Player.pan;
		Camera.rot.vx = -Player.til;
		
		
		
		// Prepare for rendering
		PrepDisplay();
		
		
		// Calculate the camera and viewpoint matrix
		CalculateCamera();
		
		
		// Set the light source
		pslt[0].vx = -(bulb_pos.vx);
		pslt[0].vy = -(bulb_pos.vy);
		pslt[0].vz = -(bulb_pos.vz);
		pslt[0].r =	0x7f;	pslt[0].g = 0x7f;	pslt[0].b = 0x7f;
		GsSetFlatLight(0, &pslt[0]);
		
		
		
		// Rotate the object
		obj_rot.vx += 8;
		obj_rot.vy += 8;
		obj_rot.vz += 8;
		
		
		// Sort the objects
		PutObject(plat_pos, plat_rot, &Object[0]);
		PutToonObject(obj_pos, obj_rot, &Object[1]);		
		PutToonObjectOut(obj_pos, obj_rot, &Object[2]);
		
		
		// Display the new frame
		Display();
		
	}
	
}


void CalculateCamera() {
	
	// This function simply calculates the viewpoint matrix based on the camera coordinates...
	// It must be called on every frame before drawing any objects.
	
	VECTOR	vec;
	GsVIEW2 view;
	
	// Copy the camera (base) matrix for the viewpoint matrix
	view.view = Camera.coord2.coord;
	view.super = WORLD;
	
	// I really can't explain how this works but I found it in one of the ZIMEN examples
	RotMatrix(&Camera.rot, &view.view);
	ApplyMatrixLV(&view.view, &Camera.pos, &vec);
	TransMatrix(&view.view, &vec);
	
	// Set the viewpoint matrix to the GTE
	GsSetView2(&view);
	
}

void PutToonObjectOut(VECTOR pos, SVECTOR rot, GsDOBJ2 *obj) {

	// Slightly modified PutObject() function for toon shading (outline stage).
	
	MATRIX lmtx,omtx;
	GsCOORDINATE2 coord;
	
	VECTOR scalev = { ONE+256, ONE+256, ONE+256};
	
	
	// First render stage (outer)
	
	// Copy the camera (base) matrix for the model
	coord = Camera.coord2;
	
	// Rotate and translate the matrix according to the specified coordinates
	ScaleMatrix(&coord.coord, &scalev);	
	RotMatrix(&rot, &omtx);
	TransMatrix(&omtx, &pos);
	CompMatrixLV(&coord.coord, &omtx, &coord.coord);
	coord.flg = 0;
	
	// Apply coordinate matrix to the object
	obj->coord2 = &coord;
	
	// Calculate Local-World (for lighting) and Local-Screen (for projection) matrices and set both to the GTE
	GsGetLws(obj->coord2, &lmtx, &omtx);
	GsSetLightMatrix(&lmtx);
	GsSetLsMatrix(&omtx);
	
	// Apply matrix to the object and then sort it!
	obj->coord2 = &coord;
	GsSortObject4(obj, &myOT[myActiveBuff], 14-OT_LENGTH, getScratchAddr(0));

	
	// Second render stage (inner)
	
	scalev.vx = scalev.vy = scalev.vz = ONE-512;
	ScaleMatrix(&coord.coord, &scalev);
	coord.flg = 0;
	
	// Calculate Local-World (for lighting) and Local-Screen (for projection) matrices and set both to the GTE
	GsGetLws(obj->coord2, &lmtx, &omtx);
	GsSetLightMatrix(&lmtx);
	GsSetLsMatrix(&omtx);
	
	// Sort the object!
	GsSortObject4(obj, &myOT[myActiveBuff], 14-OT_LENGTH, getScratchAddr(0));
	
	
}
void PutToonObject(VECTOR pos, SVECTOR rot, GsDOBJ2 *obj) {
	
	// Slightly modified PutObject() function for toon shading (fill stage).
	
	MATRIX lmtx,omtx;
	GsCOORDINATE2 coord;
	
	// Copy the camera (base) matrix for the model
	coord = Camera.coord2;
	
	// Rotate and translate the matrix according to the specified coordinates
	RotMatrix(&rot, &omtx);
	TransMatrix(&omtx, &pos);
	CompMatrixLV(&Camera.coord2.coord, &omtx, &coord.coord);
	coord.flg = 0;
	
	// Apply coordinate matrix to the object
	obj->coord2 = &coord;
	
	// Calculate Local-World (for lighting) and Local-Screen (for projection) matrices and set both to the GTE
	GsGetLws(obj->coord2, &lmtx, &omtx);
	GsSetLightMatrix(&lmtx);
	GsSetLsMatrix(&omtx);
	
	// Sort the object! (with the OT length lowered by 1 step so it'll stay above the black outline)
	GsSortObject4(obj, &myOT[myActiveBuff], 14-(OT_LENGTH-1), getScratchAddr(0));
	
}
void PutObject(VECTOR pos, SVECTOR rot, GsDOBJ2 *obj) {
	
	/*	This function draws (or sorts) a TMD model linked to a GsDOBJ2 structure... All
		matrix calculations are done automatically for simplified object placement.
		
		Parameters:
			pos 	- Object position.
			rot		- Object orientation.
			*obj	- Pointer to a GsDOBJ2 structure that is linked to a TMD model.
			
	*/
	
	MATRIX lmtx,omtx;
	GsCOORDINATE2 coord;
	
	// Copy the camera (base) matrix for the model
	coord = Camera.coord2;
	
	// Rotate and translate the matrix according to the specified coordinates
	RotMatrix(&rot, &omtx);
	TransMatrix(&omtx, &pos);
	CompMatrixLV(&Camera.coord2.coord, &omtx, &coord.coord);
	coord.flg = 0;
	
	// Apply coordinate matrix to the object
	obj->coord2 = &coord;
	
	// Calculate Local-World (for lighting) and Local-Screen (for projection) matrices and set both to the GTE
	GsGetLws(obj->coord2, &lmtx, &omtx);
	GsSetLightMatrix(&lmtx);
	GsSetLsMatrix(&omtx);
	
	// Sort the object!
	GsSortObject4(obj, &myOT[myActiveBuff], 14-OT_LENGTH, getScratchAddr(0));
	
}

int LinkModel(u_long *tmd, GsDOBJ2 *obj) {
	
	/*	This function prepares the specified TMD model for drawing and then
		links it to a GsDOBJ2 structure so it can be drawn using GsSortObject4().
		
		By default, light source calculation is disabled but can be re-enabled by
		simply setting the attribute variable in your GsDOBJ2 structure to 0.
		
		Parameters:
			*tmd - Pointer to a TMD model file loaded in memory.
			*obj - Pointer to an empty GsDOBJ2 structure.
	
		Returns:
			Number of objects found inside the TMD file.
			
	*/
	
	u_long *dop;
	int i,NumObj;
	
	// Copy pointer to TMD file so that the original pointer won't get destroyed
	dop = tmd;
	
	// Skip header and then remap the addresses inside the TMD file
	dop++; GsMapModelingData(dop);
	
	// Get object count
	dop++; NumObj = *dop;

	// Link object handler with the specified TMD
	dop++;
	for(i=0; i<NumObj; i++) {
		GsLinkObject4((u_long)dop, &obj[i], i);
		obj[i].attribute = (1<<6);	// Disables light source calculation
	}
	
	// Return the object count found inside the TMD
	return(NumObj);
	
}
void LoadTexture(u_long *addr) {
	
	// A simple TIM loader... Not much to explain
	
	RECT rect;
	GsIMAGE tim;
	
	// Get TIM information
	GsGetTimInfo((addr+1), &tim);
	
	// Load the texture image
	rect.x = tim.px;	rect.y = tim.py;
	rect.w = tim.pw;	rect.h = tim.ph;
	LoadImage(&rect, tim.pixel);
	DrawSync(0);
	
	// Load the CLUT (if present)
	if ((tim.pmode>>3) & 0x01) {
		rect.x = tim.cx;	rect.y = tim.cy;
		rect.w = tim.cw;	rect.h = tim.ch;
		LoadImage(&rect, tim.clut);
		DrawSync(0);
	}
	
}

void init() {
	
	//SVECTOR VScale={0};
	
	// Initialize the GS
	ResetGraph(0);
	if (SCREEN_YRES <= 240) {
		GsInitGraph(SCREEN_XRES, SCREEN_YRES, GsOFSGPU|GsNONINTER, DITHER, 0);
		GsDefDispBuff(0, 0, 0, 256);
	} else {
		GsInitGraph(SCREEN_XRES, SCREEN_YRES, GsOFSGPU|GsINTER, DITHER, 0);	
		GsDefDispBuff(0, 0, 0, 0);
	}
	
	// Prepare the ordering tables
	myOT[0].length	=OT_LENGTH;
	myOT[1].length	=OT_LENGTH;
	myOT[0].org		=myOT_TAG[0];
	myOT[1].org		=myOT_TAG[1];
	
	GsClearOt(0, 0, &myOT[0]);
	GsClearOt(0, 0, &myOT[1]);
	
	
	// Initialize debug font stream
	FntLoad(960, 0);
	FntOpen(-CENTERX, -CENTERY, SCREEN_XRES, SCREEN_YRES, 0, 512);
	
	
	// Setup 3D and projection matrix
	GsInit3D();
	GsSetProjection(CENTERX);
	
	// Initialize coordinates for the camera (it will be used as a base for future matrix calculations)
	GsInitCoordinate2(WORLD, &Camera.coord2);
	
	
	// Set ambient color (for lighting)
	GsSetAmbient(ONE/2, ONE/2, ONE/2);
	
	// Set default lighting mode
	GsSetLightMode(0);
	
	
	// Initialize controller
	PadInit(0);
	
}

void PrepDisplay() {
	
	// Get active buffer ID and clear the OT to be processed for the next frame
	myActiveBuff = GsGetActiveBuff();
	GsSetWorkBase((PACKET*)myPacketArea[myActiveBuff]);
	GsClearOt(0, 0, &myOT[myActiveBuff]);
	
}
void Display() {
	
	// Flush the font stream
	FntFlush(-1);
	
	// Wait for VSync, switch buffers, and draw the new frame.
	VSync(0);
	GsSwapDispBuff();
	GsSortClear(0, 64, 0, &myOT[myActiveBuff]);
	GsDrawOt(&myOT[myActiveBuff]);
	
}